#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <cstring>
#include <algorithm>

#include <sdsl/bit_vectors.hpp>
#include <sdsl/hybrid_vector.hpp>

int random_int(int p, int r) {
  return p + rand() % (r - p + 1);
}

void test_random(int length, int sigma, int density, char keybit, int strings, int queries) {
  std::cout << "TEST, length = " << length << ", "
    << "sigma = " << sigma << ", "
    << "density = " << density << ", "
    << "keybit = " << (char)('0' + keybit) << ", "
    << "strings = " << strings << ", "
    << "queries  = " << queries << "\n";
  for (int strings_ptr = 0; strings_ptr < strings; ++strings_ptr) {
    // Create random bitvector.
    sdsl::bit_vector v(length);
    for (int i = 0; i < length; ++i) {
      if (random_int(0, density - 1) == 0) v[i] = keybit;
      else v[i] = 1 - keybit;
    }

    // Initialize hybrid bitvector.
    sdsl::hybrid_vector<16> hybrid_v(v); // hybrid bitvector
    sdsl::hybrid_vector<16>::rank_1_type rank_hybrid_v(&hybrid_v); // rank support

    // Initialize other bitvector for checking.
    sdsl::rank_support_v5<> rank_v(&v); // rank support for uncompr. bitvector
    sdsl::rrr_vector<15> rrr_v(v); // RRR bitvector 
    sdsl::rrr_vector<15>::rank_1_type rank_rrr_v(&rrr_v); // rank supporst for RRR

    for (int q = 1; q <= queries; ++q) {
      int i = random_int(0, length - 1);

      int ret_hybrid = rank_hybrid_v.rank(i);
      int ret_uncompr = rank_v.rank(i);
      int ret_rrr = rank_rrr_v.rank(i);

      if (ret_hybrid != ret_uncompr || ret_uncompr != ret_rrr) {
        std::cerr << "Error\n";
        std::cerr << "\tlength = " << length << ", "
          << "q = " << q << ", "
          << "i = " << i << "\n";
        std::cerr << "\tret_hybrid = " << ret_hybrid << ", "
          << "ret_uncompr = " << ret_uncompr << ", "
          << "ret_rrr = " << ret_rrr << "\n";

        if (length < 1000)
          std::cerr << "\tbitvector = " << v << "\n";
          
        std::exit(EXIT_FAILURE);
      }
    }
  }
}

int main() {
  srand(time(0) + getpid());
  fprintf(stderr, "Testing hybrid bitvector rank queries on random data.\n");

  for (int length = 1; length <= (1 << 20); length *= 2) {
    for (int density = 1; density <= 32; density *= 2) {
      test_random(length, 2, density, 1, 50, std::min(20000, 5 * length));
      test_random(length, 2, density, 0, 50, std::min(20000, 5 * length));
     }
  }

  fprintf(stderr,"All tests passed.\n");
}

